/*
 * User-space  module driver
 *
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <linux/clk.h>
#include <linux/module.h>
#include <linux/of_gpio.h>
#include <linux/miscdevice.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/virtio.h>
#include <linux/uio_driver.h>
#include <linux/uaccess.h>

#define DRIVER_NAME		"uio_module_drv"
#define DRIVER_VERSION		"0.1"

/**
 * struct uio_module_drv_addr_map - uio module driver address map entry
 */
struct uio_module_drv_addr_map {
	phys_addr_t	addr;
	unsigned long	length;
	int		index;
	struct list_head	node;
};

/**
 * struct uio_module_drv_info - local information for uio module driver
 */
struct uio_module_drv_info {
	struct uio_info		uio;
	struct device		*dev;
	struct miscdevice	misc;
	struct clk		*clk;
	unsigned long		flags;
	spinlock_t		lock;
	int			start_offset;
	struct list_head	addr_map;
	struct work_struct	workqueue;
};

/**
 * uio_module_drv_open - uio module driver open routine
 */
static int uio_module_drv_open(struct uio_info *uio, struct inode *inode)
{
	struct uio_module_drv_info *uio_module_drv = uio->priv;
	if (uio_module_drv->clk)	
		return clk_prepare_enable(uio_module_drv->clk);
	else
		return(0);
}

/**
 * uio_module_drv_release - uio module driver release routine
 */
static int uio_module_drv_release(struct uio_info *uio, struct inode *inode)
{
	struct uio_module_drv_info *uio_module_drv = uio->priv;

	if (uio_module_drv->clk)	
		clk_disable_unprepare(uio_module_drv->clk);
	return 0;
}

/**
 * uio_module_drv_handler - uio module driver interrupt handler
 */
static irqreturn_t uio_module_drv_handler(int irq, struct uio_info *uio)
{
	struct uio_module_drv_info *uio_module_drv = uio->priv;

	if (!test_and_set_bit(0, &uio_module_drv->flags))
		disable_irq_nosync(irq);

	return IRQ_HANDLED;
}

/**
 * uio_module_drv_irqcontrol - uio module driver interrupt control
 */
static int uio_module_drv_irqcontrol(struct uio_info *uio, s32 irq_on)
{
	struct uio_module_drv_info *uio_module_drv = uio->priv;
	unsigned long flags;

	spin_lock_irqsave(&uio_module_drv->lock, flags);
	if (irq_on) {
		if (test_and_clear_bit(0, &uio_module_drv->flags))
			enable_irq(uio->irq);
	} else {
		if (!test_and_set_bit(0, &uio_module_drv->flags))
			disable_irq(uio->irq);
	}
	spin_unlock_irqrestore(&uio_module_drv->lock, flags);

	return 0;
}

/**
 * uio_module_drv_dev_fop_open() : open for the uio module driver
 */
static int uio_module_drv_dev_fop_open(struct inode *inode, struct file *file)
{
	/* Need an empty open so that file->private_data gets populated */
	return 0;
}

/**
 * uio_module_drv_dev_fop_mmap() - provided mmap support for
 * device memory. This checks if user request is in valid range before providing
 * mmap access. The valid range can be configured using device tree or platform
 * data.
 */
static int uio_module_drv_dev_fop_mmap(struct file *file, struct vm_area_struct *vma)
{
	bool is_allowed = false;
	struct uio_module_drv_addr_map *pmap;
	size_t size = vma->vm_end - vma->vm_start;
	phys_addr_t addr = vma->vm_pgoff << PAGE_SHIFT;
	struct miscdevice *misc = file->private_data;
	struct uio_module_drv_info *uio_module_drv =
		container_of(misc, struct uio_module_drv_info, misc);

	/* Check if the requested address and length is in the
	range configured during probe */
	list_for_each_entry(pmap, &uio_module_drv->addr_map, node) {
		if ((pmap->addr & PAGE_MASK)<= addr &&
		((((pmap->addr + pmap->length) & PAGE_MASK) + PAGE_SIZE)
		 >= ((addr + size))))  {
				is_allowed = true;
				break;
		}
	}

	if (!is_allowed) {
		dev_err(uio_module_drv->dev,
			"mmap offset 0x%lx is outside the allowed range\n",
			(long unsigned int) addr);
		return -EINVAL;
	}

	if (vma->vm_start & ~PAGE_MASK) {
		dev_err(uio_module_drv->dev, "must mmap at page boundary\n");
		return -EINVAL;
	}

	vma->vm_page_prot = phys_mem_access_prot(file, vma->vm_pgoff, size,
						 vma->vm_page_prot);

	if (remap_pfn_range(vma,
			    vma->vm_start,
			    vma->vm_pgoff,
			    size,
			    vma->vm_page_prot)) {
		return -EAGAIN;
	}
	return 0;

}
/**
 * uio_module_drv_transfer() - provides ability to write to memory areas
 */

ssize_t  uio_module_drv_transfer(struct file *file, char __user *buf,
			size_t count, loff_t *ppos, int flag)
{
	bool is_allowed = false;
	struct uio_module_drv_addr_map *pmap;
	struct miscdevice *misc = file->private_data;
	struct uio_module_drv_info *uio_module_drv =
		container_of(misc, struct uio_module_drv_info, misc);
	uint32_t phys_addr = *(ppos);
	uint32_t *virt_addr_p=NULL;

	/* Check if the requested address and length is in the
	range configured duing probe */
	list_for_each_entry(pmap, &uio_module_drv->addr_map, node) {
		if (pmap->addr <= phys_addr &&
			pmap->addr + pmap->length >= phys_addr + count) {
				is_allowed = true;
				break;
		}
	}

	if (!is_allowed) {
		dev_err(uio_module_drv->dev,
			"offset 0x%lx is outside the allowed range\n",
			(long unsigned int) phys_addr);
		return -EINVAL;
	}
	virt_addr_p = ioremap_nocache (phys_addr, count);
	if (virt_addr_p == NULL) {
		dev_err(uio_module_drv->dev, "Mapping of virtual memory failed\n");
		return(-ENOMEM);
	}
	if (flag == 1) {
		if(copy_from_user(virt_addr_p, buf, count)) {
			dev_err(uio_module_drv->dev, "copy_from_user failed\n");
			return -EFAULT;
		}
	}
	else {
		if(copy_to_user(buf, virt_addr_p, count)) {
			dev_err(uio_module_drv->dev, "copy_to_user failed\n");
			return -EFAULT;
		}
	}
	iounmap(virt_addr_p);
	return(count);
	
}
/**
 * uio_module_drv_dev_fop_write() - provides ability to write to memory areas
 */

ssize_t  uio_module_drv_dev_fop_write(struct file *file, const char __user *buf,
			size_t count, loff_t *ppos)
{
	return(uio_module_drv_transfer(file, (char __user *)buf, count, ppos, 1));
}



/**
 * uio_module_drv_dev_fop_read() - provides ability to read from memory areas
 */

ssize_t  uio_module_drv_dev_fop_read(struct file *file, char __user *buf,
			size_t count, loff_t *ppos)
{
	printk("\nuio_read: Device read entered : %p : %x", ppos, (uint32_t)(*ppos));
	return(uio_module_drv_transfer(file, buf, count, ppos, 0));
}

static const struct file_operations uio_module_drv_dev_fops = {
	.owner		= THIS_MODULE,
	.open		= uio_module_drv_dev_fop_open,
	.mmap		= uio_module_drv_dev_fop_mmap,
	.read           = uio_module_drv_dev_fop_read,
	.write          = uio_module_drv_dev_fop_write,
};

/**
 * uio_module_drv_populate_segments() - scan the configuration for "mem" and populate
 * polulate in uio_module_drv local structure. This information will be used to process
 * user mmap requests.
 */
#ifdef CONFIG_OF
static inline int uio_module_drv_populate_segments(struct device_node *np,
					struct uio_module_drv_info *uio_module_drv)
{
	int len, i;
	u32 *paddr_map;
	int num_maps = 0;
	struct uio_module_drv_addr_map *pmapentry;

	if (of_get_property(np, "mem", &len)) {

		paddr_map = devm_kzalloc(uio_module_drv->dev, len, GFP_KERNEL);
		if (!paddr_map) {
			dev_err(uio_module_drv->dev, "memory allocation failed\n");
			return -ENOMEM;
		}
		/* check if length even multiple of sizeof(u32), i.e.,
		    the dt bindings need to be of the form <addr length>
		 */
		len = len / sizeof(u32);
		if ((len % 2) != 0) {
			dev_err(uio_module_drv->dev, "invalid address map in dt binding\n");
			return -EINVAL;
		}
		num_maps = len / 2;
		if (of_property_read_u32_array(np, "mem",
					       paddr_map, len)) {
			dev_err(uio_module_drv->dev, "No addr-map array  in dt bindings\n");
			return -ENODEV;
		}

		pmapentry  = devm_kzalloc(uio_module_drv->dev,
				sizeof(struct uio_module_drv_addr_map) * num_maps,
				GFP_KERNEL);
		if (!pmapentry) {
			dev_err(uio_module_drv->dev, "devm_kzalloc mapping failed\n");
			return -ENOMEM;
		}

		/* populate the uio_module_drv structure for policing */
		for (i = 0; i < num_maps; i++) {
			pmapentry->index = i;
			pmapentry->addr = *paddr_map++;
			pmapentry->length = *paddr_map++;
			list_add_tail(&pmapentry->node, &uio_module_drv->addr_map);
			pmapentry++;
		}

	}
	return 0;
}
#else
static inline void uio_module_drv_populate_segments(struct device_node *np,
					struct uio_module_drv_info *uio_module_drv)
{
	return 0;
}
#endif

/**
 * uio_module_drv_driver_probe() probe routine for the uio module driver
 */
static int uio_module_drv_driver_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct device_node *np = dev->of_node;
	struct uio_module_drv_info *uio_module_drv;
	struct miscdevice *misc;
	struct uio_info *uio;
	struct resource *r;
	int error = 0;
	int i;
	int irq;
	
	uio_module_drv = kzalloc(sizeof(struct  uio_module_drv_info), GFP_KERNEL);
	if (!uio_module_drv) {
		dev_err(dev, "%s: kzalloc failed\n", __func__);
		error = -ENOMEM;
		goto fail;
	}

	uio_module_drv->dev	= dev;
	spin_lock_init(&uio_module_drv->lock);

	uio		= &uio_module_drv->uio;
	uio->name	= dev_name(dev);
	uio->version	= DRIVER_VERSION;
	uio->priv	= uio_module_drv;
	uio->handler	= uio_module_drv_handler;
	uio->irqcontrol	= uio_module_drv_irqcontrol;
	uio->open	= uio_module_drv_open;
	uio->release	= uio_module_drv_release;

	irq	= platform_get_irq(pdev, 0);
	if (!(irq < 0))
		/* pass up control irq to user-space */
		uio->irq	= irq;

	for (i = 0; i < MAX_UIO_MAPS; ++i) {
		r = platform_get_resource(pdev, IORESOURCE_MEM, i);
		if (!r)
			break;
		uio->mem[i].memtype = UIO_MEM_PHYS;
		uio->mem[i].addr    = r->start;
		uio->mem[i].size    = resource_size(r);
		uio->mem[i].name    = r->name;
	}

	uio_module_drv->clk = clk_get(dev, NULL);
	if(IS_ERR(uio_module_drv->clk))
		uio_module_drv->clk = NULL;

	INIT_LIST_HEAD(&uio_module_drv->addr_map);
	error = uio_module_drv_populate_segments(np, uio_module_drv);
	if (error)
		goto fail_uio;
	if (uio->irq || uio->mem[0].memtype != UIO_MEM_NONE) {
		error = uio_register_device(dev, uio);
		if (error) {
			dev_err(dev, "failed to register uio device\n");
			goto fail_uio;
		}
	}
	platform_set_drvdata(pdev, uio_module_drv);

	if (!list_empty(&uio_module_drv->addr_map)) {
		misc = &uio_module_drv->misc;

		misc->minor	= MISC_DYNAMIC_MINOR;
#ifdef CONFIG_OF
		if (of_property_read_string(np, "label", &misc->name) < 0)
			misc->name = dev_name(dev);
		if (!misc->name)
			misc->name = "unknown";
#else
		misc->name	= dev_name(dev);
#endif
		misc->fops	= &uio_module_drv_dev_fops;
		misc->parent	= dev;

		if (misc_register(misc)) {
			dev_err(dev, "could not register misc device\n");
			goto fail_uio_module;
		}
		dev_info(dev, "registered misc device %s\n", misc->name);
	}

	return 0;

fail_uio_module:
	uio_unregister_device(uio);
fail_uio:
	if(uio_module_drv->clk)
		clk_put(uio_module_drv->clk);
	devm_kfree(dev, uio_module_drv);
fail:
	return error;
}
/**
 * uio_module_drv_driver_remove() remove routine for the uio module driver
 */

static int uio_module_drv_driver_remove(struct platform_device *pdev)
{
	struct uio_module_drv_info *uio_module_drv = platform_get_drvdata(pdev);
	struct uio_info *uio = &uio_module_drv->uio;

	if (uio_module_drv) {
		if (uio->irq || uio->mem[0].memtype != UIO_MEM_NONE)
			uio_unregister_device(&uio_module_drv->uio);
		if(uio_module_drv->clk)
			clk_put(uio_module_drv->clk);
		if (!list_empty(&uio_module_drv->addr_map))
			misc_deregister(&uio_module_drv->misc);
	}
	platform_set_drvdata(pdev, NULL);
	return 0;
}

static struct of_device_id uio_module_drv_of_match[] = {
	{ .compatible = "ti,uio-module-drv", },
	{},
};
MODULE_DEVICE_TABLE(of, uio_module_drv_of_match);

static struct platform_driver uio_module_drv_driver = {
	.driver	= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = uio_module_drv_of_match,
	},
	.probe	= uio_module_drv_driver_probe,
	.remove	= uio_module_drv_driver_remove,
};

module_platform_driver(uio_module_drv_driver);
MODULE_AUTHOR("Sam Nelson");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("User-space driver for a generic module ");
MODULE_ALIAS("platform:" DRIVER_NAME);
